from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict, List, Tuple, Any, Optional, Callable, Union, Iterable, Protocol, TypeVar

if TYPE_CHECKING:
    import threading
    from concurrent.futures import Future

    from .sysinfo import SysInfo
    from .task import Task
    from .worker import Worker


T = TypeVar('T')

class QueueLike(Protocol[T]):
    def put(self, item: T) -> None: ...
    def get(self) -> T: ...


class Pool(ABC):

    __sys_info_lock:Optional[threading.Lock] = None
    __sys_info:Optional[SysInfo] = None

    def __init__(
        self, max_workers:int,

        initializer:Optional[Callable[..., Any]],
        initargs:Tuple[Any, ...],
        initkwargs:Optional[Dict[str, Any]],

        result_queue_cls:type,
        result_queue_args:Tuple[Any, ...]=(),
        result_queue_kwargs:Dict[str, Any]=None,
        
        *,

        max_tasks_per_child:Optional[int],
        use_torch:bool,
        need_module_deps:bool
    ):
        import threading
        import os
        

        if use_torch:
            import torch
            self._torch_cuda_available = torch.cuda.is_available()
        else:
            self._torch_cuda_available = False


        if not max_workers:
            max_workers = os.cpu_count()

        self._max_tasks_per_child:int = max_tasks_per_child
        self._need_module_deps:bool = need_module_deps
        self._use_torch:bool = use_torch
        self._max_workers:int = max_workers
        self._initializer:Optional[Callable[..., Any]] = initializer
        self._initargs:Tuple[Any, ...] = initargs
        self._initkwargs:Optional[Dict[str, Any]] = initkwargs

        if result_queue_kwargs is None:
            result_queue_kwargs = {}

        self._result_queue:QueueLike[Tuple[str, bool, Any]] = result_queue_cls(*result_queue_args, **result_queue_kwargs)
        self._workers:List[Worker] = []
        self._tasks:Dict[str, Task] = {}
        self._delayed_tasks:List[Task] = []
        self._lock = threading.Lock()
        self._shutdown = False
        self._result_thread = None

    def submit(
        self, func:Callable[..., Any],
        args:Optional[Tuple[Any]]=None, kwargs:Optional[Dict[str, Any]]=None,
        need_cpu_cores:int=1, need_cpu_mem:int=0,
        need_gpu_cores:int=0, need_gpu_mem:int=0
    )->Future:
        import threading
        from .task import Task

        if args is None:
            args = []

        if kwargs is None:
            kwargs = {}

        with self._lock:
            if self._shutdown:
                raise RuntimeError("cannot submit after shutdown")
            
            task = Task(
                func=func,
                args=args,
                kwargs=kwargs,
                need_cpu_cores=need_cpu_cores,
                need_cpu_mem=need_cpu_mem,
                need_gpu_cores=need_gpu_cores,
                need_gpu_mem=need_gpu_mem,
                calculate_module_deps=self._need_module_deps
            )
            self._tasks[task.id] = task
            
            worker = self._choose_task_worker(task)
            if worker is None:
                self._delayed_tasks.append(task)
                return task.future

            device = self._choose_task_device(task)
            if device is None:
                self._delayed_tasks.append(task)
                return task.future
            
            self._put_task(task)
            if self._result_thread is None:
                self._result_thread = threading.Thread(target=self._collecting_result, daemon=True, name="collecting_result")
                self._result_thread.start()

            return task.future
        
    @property
    def _sys_info(self)->SysInfo:
        from .sysinfo import SysInfo
        
        if Pool.__sys_info is None:
            Pool.__sys_info = SysInfo()

        return Pool.__sys_info
    
    @property
    def _sys_info_lock(self)->threading.Lock:
        import threading

        if Pool.__sys_info_lock is None:
            Pool.__sys_info_lock = threading.Lock()

        return Pool.__sys_info_lock

    @staticmethod
    def _result_iterator(futures:List[Future], end_time:Optional[float]):
        from concurrent.futures._base import _result_or_cancel
        import time

        try:
            futures.reverse()
            while futures:
                if end_time is None:
                    yield _result_or_cancel(futures.pop())
                else:
                    yield _result_or_cancel(futures.pop(), end_time - time.monotonic())
        finally:
            for future in futures:
                future.cancel()

    def starmap(
        self, func:Callable[..., Any],
        args_iterables:Iterable[Tuple[Any, ...]],
        need_cpu_cores:Union[int, Iterable[int]]=1,
        need_cpu_mem:Union[int, Iterable[int]]=0,
        need_gpu_cores:Union[int, Iterable[int]]=0,
        need_gpu_mem:Union[int, Iterable[int]]=0,
        timeout:Optional[Union[float, int]]=None,
        chunksize:int=1
    )->Iterable[Any]:
        from functools import partial
        import itertools
        import time
        from collections.abc import Iterable
        from concurrent.futures.process import _process_chunk, _chain_from_iterable_of_lists
        from .utils import batched

        if not isinstance(need_cpu_cores, Iterable):
            need_cpu_cores = itertools.repeat(need_cpu_cores)

        if not isinstance(need_cpu_mem, Iterable):
            need_cpu_mem = itertools.repeat(need_cpu_mem)

        if not isinstance(need_gpu_cores, Iterable):
            need_gpu_cores = itertools.repeat(need_gpu_cores)

        if not isinstance(need_gpu_mem, Iterable):
            need_gpu_mem = itertools.repeat(need_gpu_mem)

        if chunksize < 1:
            raise ValueError("chunksize must be >= 1.")

        end_time = None
        if timeout is not None:
            end_time = timeout + time.monotonic()

        target_func = partial(_process_chunk, func)
        futures:List[Future] = []
        iterator = zip(args_iterables, need_cpu_cores, need_cpu_mem, need_gpu_cores, need_gpu_mem)
        for batch in batched(iterator, chunksize):
            args_batch, cpu_cores_batch, cpu_mem_batch, gpu_cores_batch, gpu_mem_batch = zip(*batch)
            future = self.submit(
                target_func, args=(args_batch,),
                need_cpu_cores=max(cpu_cores_batch),
                need_cpu_mem=max(cpu_mem_batch),
                need_gpu_cores=max(gpu_cores_batch),
                need_gpu_mem=max(gpu_mem_batch)
            )
            futures.append(future)

        return _chain_from_iterable_of_lists(Pool._result_iterator(futures, end_time))

    def map(
        self, func:Callable[..., Any],
        iterable:Iterable[Any],
        need_cpu_cores:Union[int, Iterable[int]]=1,
        need_cpu_mem:Union[int, Iterable[int]]=0,
        need_gpu_cores:Union[int, Iterable[int]]=0,
        need_gpu_mem:Union[int, Iterable[int]]=0,
        timeout:Optional[Union[float, int]]=None,
        chunksize:int=1
    )->Iterable[Any]:
        args_iterable = ((item,) for item in iterable)
        
        return self.starmap(
            func=func,
            args_iterables=args_iterable,
            need_cpu_cores=need_cpu_cores,
            need_cpu_mem=need_cpu_mem,
            need_gpu_cores=need_gpu_cores,
            need_gpu_mem=need_gpu_mem,
            timeout=timeout,
            chunksize=chunksize
        )

    def shutdown(self, wait:bool=True, *, cancel_futures:bool=False)->None:
        with self._lock:
            if self._shutdown:
                return
            
            self._shutdown = True

            for worker in self._workers:
                worker.stop()

            if cancel_futures:
                for task in self._tasks.values():
                    task.future.cancel()

            if wait:
                for worker in self._workers:
                    worker.join()

    def __enter__(self)->Pool:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb)->None:
        self.shutdown(wait=True)

    def _collecting_result(self)->None:
        while not self._shutdown:
            task_id, success, result = self._result_queue.get()

            with self._lock:
                task = self._tasks.pop(task_id)
                if success:
                    task.future.set_result(result)
                else:
                    task.future.set_exception(result)
                
                worker:Worker = task.worker
                worker.is_working = False
                worker.n_finished_tasks += 1
                if self._max_tasks_per_child is not None and worker.n_finished_tasks >= self._max_tasks_per_child:
                    worker.restart()

                self._release_resource(task)
                self._postprocess_after_task_done()
    
    def _postprocess_after_task_done(self):
        should_pop_indices = []
        cancelled_task_ids = []
        for i, delayed_task in enumerate(self._delayed_tasks):
            if delayed_task.future.cancelled():
                cancelled_task_ids.append(delayed_task.id)
                should_pop_indices.append(i)
                continue

            used_worker = self._choose_task_worker(delayed_task)
            if used_worker is None:
                continue

            device = self._choose_task_device(delayed_task)
            if device is None:
                continue

            self._put_task(delayed_task)
            should_pop_indices.append(i)

        for i in reversed(should_pop_indices):
            del self._delayed_tasks[i]

        for task_id in cancelled_task_ids:
            del self._tasks[task_id] 

        if self._torch_cuda_available:
            for task in self._tasks.values():
                self._try_move_to_gpu(task)

    def _choose_task_device(self, task:Task)->Optional[str]:
        from .worker import Worker

        with self._sys_info_lock:
            if Worker.total_working_count() == 0:
                self._sys_info.update()

            need_cpu_cores = self._estimate_need_cpu_cores(task)
            if need_cpu_cores > self._sys_info.cpu_cores_free:
                task.device = None
                task.worker = None
                return None

            task.estimated_need_cpu_mem = self._estimate_need_cpu_mem(task)
            if task.estimated_need_cpu_mem > self._sys_info.cpu_mem_free:
                task.device = None
                task.worker = None
                return None

            if not self._torch_cuda_available or (task.need_gpu_cores == 0 and task.need_gpu_mem == 0):
                task.device = "cpu"
                return "cpu"

            gpus = self._sys_info.gpu_infos
            if not gpus:
                task.device = "cpu"
                return "cpu"

            best_gpu = None
            for gpu in gpus:
                need_gpu_cores:int = self._estimate_need_gpu_cores(task, gpu.id)
                if gpu.mem_free >= task.need_gpu_mem and gpu.n_cores_free >= need_gpu_cores:
                    if best_gpu is None or gpu.n_cores_free > best_gpu.n_cores_free:
                        best_gpu = gpu

        if best_gpu is None:
            task.device = "cpu"
            return "cpu"

        task.device = best_gpu.device
        return task.device

    def _choose_task_worker(self, task:Task)->Optional[Worker]:
        best_worker:Optional[Worker] = None
        task.modules_overlap_ratio = 0.0
        for worker in self._workers:
            if worker.is_working:
                continue

            if not self._need_module_deps:
                task.worker = worker
                return worker

            if task.need_cpu_mem == 0 or task.func.__module__ in worker.imported_modules:
                task.modules_overlap_ratio = 1.0
                task.worker = worker
                return worker

            current_overlap_ratio = worker.overlap_modules_ratio(task)
            if best_worker is None or current_overlap_ratio > task.modules_overlap_ratio:
                task.modules_overlap_ratio = current_overlap_ratio
                best_worker = worker
            
        if best_worker is not None:
            task.worker = best_worker
            return best_worker
            
        task.modules_overlap_ratio = 0.0
        if len(self._workers) < self._max_workers:
            task.worker = self._add_worker()
        else:
            task.worker = None

        return task.worker

    @abstractmethod
    def _take_resource(self, task:Task)->None:
        pass

    @abstractmethod
    def _release_resource(self, task:Task)->None:
        pass

    @abstractmethod
    def _estimate_need_gpu_cores(self, task:Task, gpu_id:int)->int:
        pass

    @abstractmethod
    def _estimate_need_cpu_cores(self, task:Task)->int:
        pass
    
    @abstractmethod
    def _estimate_need_cpu_mem(self, task:Task)->int:
        pass

    @abstractmethod
    def _put_task(self, task:Task)->None:
        pass

    def _try_move_to_gpu(self, task:Task)->None:
        if (
            task.device is None or
            task.device.startswith("cuda") or
            task.need_gpu_cores == 0 or
            task.worker is None or
            not task.worker.is_working
        ):
            return
        
        with self._sys_info_lock:
            gpus = self._sys_info.gpu_infos
            if not gpus:
                return
            
            need_gpu_mem:int = task.need_gpu_mem

            best_gpu = None
            need_best_gpu_cores:int = 0
            for gpu in gpus:
                need_gpu_cores:int = self._estimate_need_gpu_cores(task, gpu.id)
                if gpu.mem_free >= need_gpu_mem and gpu.n_cores_free >= need_gpu_cores:
                    if best_gpu is None or gpu.n_cores_free > best_gpu.n_cores_free:
                        best_gpu = gpu
                        need_best_gpu_cores = need_gpu_cores

            if best_gpu is None:
                return

            worker:Worker = task.worker
            worker.change_device(best_gpu.device)
            task.device = best_gpu.device
            best_gpu.n_cores_free -= need_best_gpu_cores
            best_gpu.mem_free -= task.need_gpu_mem

    @abstractmethod
    def _add_worker(self)->Worker:
        pass


from .processpool import ProcessPool
from .threadpool import ThreadPool
from .interpreterpool import InterpreterPool
from .utils import DataSize, move_optimizer_to, limit_num_single_thread, best_device



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional, Callable, Any, Dict, Tuple, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from queue import Queue
    from .task import Task


class Worker(ABC):

    _total_working_count:int = 0

    def __init__(
        self, index:int,
        initializer:Optional[Callable[..., Any]],
        initargs:Tuple[Any, ...],
        initkwargs:Optional[Dict[str, Any]]
    ):
        self.index:int = index
        self._is_working:bool = False
        self.initializer:Optional[Callable[..., Any]] = initializer
        self.initargs:Tuple[Any, ...] = initargs
        self.initkwargs:Optional[Dict[str, Any]] = initkwargs
        self.imported_modules:Set[str] = set()
        self.n_finished_tasks:int = 0
    
    @property
    def is_working(self)->bool:
        return self._is_working

    @is_working.setter
    def is_working(self, is_working:bool)->None:
        if self._is_working == is_working:
            return
        
        self._is_working = is_working

        if is_working:
            Worker._total_working_count += 1
        else:
            Worker._total_working_count -= 1

    @staticmethod
    def total_working_count()->int:
        return Worker._total_working_count

    @abstractmethod
    def change_device(self, device:str)->None:
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def join(self)->None:
        pass

    @abstractmethod
    def restart(self)->None:
        pass

    def overlap_modules_ratio(self, task:Task)->float:
        if not self.imported_modules:
            return 0
        
        return len(self.imported_modules & task.module_deps) / len(self.imported_modules)

    @staticmethod
    def _changing_device(cmd_queue:Queue[Optional[str]]):
        from .utils import _set_best_device
        while True:
            device = cmd_queue.get()
            _set_best_device(device)

    @staticmethod
    def run(
        task_queue:Queue[Optional[Tuple[str, Callable[..., Any], Tuple[Any, ...], Dict[str, Any]]]],
        result_queue:Queue[Optional[Tuple[str, bool, Any]]],
        change_device_cmd_queue:Optional[Queue[Optional[str]]],
        initializer:Optional[Callable[..., Any]],
        initargs:Tuple[Any, ...],
        initkwargs:Optional[Dict[str, Any]]
    ):
        from .utils import _set_best_device
        

        if initializer is not None:
            if initkwargs is None:
                initkwargs = {}

            initializer(*initargs, **initkwargs)
        
        if change_device_cmd_queue is not None:
            import threading

            change_device_thread = threading.Thread(target=Worker._changing_device, args=(change_device_cmd_queue,), daemon=True, name="changing_device")
            change_device_thread.start()

        while True:
            task = task_queue.get()
            if task is None:
                break

            task_id, task_device, func, args, kwargs = task
            _set_best_device(task_device)

            try:
                result = func(*args, **kwargs)
                success = True
            except BaseException as e:
                result = e
                success = False

            result_queue.put((task_id, success, result))


from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Tuple, Any, Optional, Callable

from ..pool import Pool

if TYPE_CHECKING:
    from ..task import Task
    from .processworker import ProcessWorker


class ProcessPool(Pool):

    def __init__(
        self, max_workers:int=0,
        process_name_prefix:str="ProcessPool.worker:",
        mp_context:str="spawn",
        initializer:Optional[Callable[..., Any]]=None,
        initargs:Tuple[Any, ...]=(),
        initkwargs:Optional[Dict[str, Any]]=None,
        *,
        max_tasks_per_child:Optional[int]=None,
        use_torch:bool=False
    ):
        import threading
        import queue

        if use_torch:
            import torch.multiprocessing as mp
            from torch.multiprocessing.queue import SimpleQueue
        else:
            import multiprocessing as mp
            from multiprocessing.queues import SimpleQueue

        self._ctx = mp.get_context(mp_context)
        Pool.__init__(
            self, max_workers=max_workers,

            initializer=initializer,
            initargs=initargs,
            initkwargs=initkwargs,

            result_queue_cls=SimpleQueue,
            result_queue_kwargs={"ctx": self._ctx},
            
            max_tasks_per_child=max_tasks_per_child,
            use_torch=use_torch,
            need_module_deps=True
        )

        self._process_name_prefix:str = process_name_prefix
        
        self._feeding_queue:queue.SimpleQueue[Tuple[Task, SimpleQueue]] = queue.SimpleQueue()
        self._feeding_thread = threading.Thread(target=self._feeding, daemon=True, name="feeding")
        self._feeding_thread.start()

    def _take_resource(self, task):
        with self._sys_info_lock:
            self._sys_info.cpu_cores_free -= task.need_cpu_cores
            self._sys_info.cpu_mem_free -= task.estimated_need_cpu_mem
            task_gpu_id:int = task.gpu_id
            if task_gpu_id != -1:
                self._sys_info.gpu_infos[task_gpu_id].n_cores_free -= task.need_gpu_cores
                self._sys_info.gpu_infos[task_gpu_id].mem_free -= task.need_gpu_mem

    def _release_resource(self, task:Task):
        with self._sys_info_lock:
            self._sys_info.cpu_cores_free += task.need_cpu_cores
            worker:ProcessWorker = task.worker
            hold_cpu_mem = max(worker.cached_rss - task.mem_before_enter, 0)
            self._sys_info.cpu_mem_free += max(task.need_cpu_mem - hold_cpu_mem, 0)
            task_gpu_id:int = task.gpu_id
            if task_gpu_id != -1:
                self._sys_info.gpu_infos[task_gpu_id].n_cores_free += task.need_gpu_cores
                self._sys_info.gpu_infos[task_gpu_id].mem_free += task.need_gpu_mem

    def _estimate_need_cpu_cores(self, task):
        return task.need_cpu_cores
    
    def _estimate_need_cpu_mem(self, task):
        worker:ProcessWorker = task.worker
        task.mem_before_enter = worker.cached_rss
        return max(0, task.need_cpu_mem - task.modules_overlap_ratio * worker.cached_rss)
    
    def _estimate_need_gpu_cores(self, task, gpu_id):
        return task.need_gpu_cores

    def _put_task(self, task:Task)->None:
        self._take_resource(task)
        worker:ProcessWorker = task.worker
        worker.is_working = True
        worker.imported_modules.update(task.module_deps)
        self._feeding_queue.put((task, worker.task_queue))
        
    def _feeding(self)->None:
        while not self._shutdown:
            task, task_queue = self._feeding_queue.get()
            if task.future.cancelled():
                continue

            try:
                task_queue.put(task.info())
                task.future.set_running_or_notify_cancel()
            except BaseException as e:
                task.future.set_exception(e)

    def _add_worker(self)->ProcessWorker:
        from .processworker import ProcessWorker

        worker = ProcessWorker(
            len(self._workers), self._process_name_prefix,
            self._result_queue, self._ctx,
            initializer=self._initializer,
            initargs=self._initargs,
            initkwargs=self._initkwargs,
            use_torch=self._use_torch,
            torch_cuda_available=self._torch_cuda_available
        )
        self._workers.append(worker)
        return worker



from __future__ import annotations
from ..pool import Pool
from typing import TYPE_CHECKING, Dict, Tuple, Any, Optional, Callable

if TYPE_CHECKING:
    from ..task import Task
    from .interpreterworker import InterpreterWorker


class InterpreterPool(Pool):

    def __init__(
        self, max_workers:int=0,
        initializer:Optional[Callable[..., Any]]=None,
        initargs:Tuple[Any, ...]=(),
        initkwargs:Optional[Dict[str, Any]]=None,
        *,
        max_tasks_per_child:Optional[int]=None,
        use_torch:bool=False
    ):
        import concurrent.interpreters as interpreters

        Pool.__init__(
            self, max_workers=max_workers,
            
            initializer=initializer,
            initargs=initargs,
            initkwargs=initkwargs,

            result_queue_cls=interpreters.create_queue,

            max_tasks_per_child=max_tasks_per_child,
            use_torch=use_torch,
            need_module_deps=True
        )

    def _take_resource(self, task):
        with self._sys_info_lock:
            self._sys_info.cpu_cores_free -= task.need_cpu_cores
            self._sys_info.cpu_mem_free -= task.estimated_need_cpu_mem
            task_gpu_id:int = task.gpu_id
            if task_gpu_id != -1:
                self._sys_info.gpu_infos[task_gpu_id].n_cores_free -= task.need_gpu_cores
                self._sys_info.gpu_infos[task_gpu_id].mem_free -= task.need_gpu_mem

    def _release_resource(self, task):
        with self._sys_info_lock:
            self._sys_info.cpu_cores_free += task.need_cpu_cores
            self._sys_info.cpu_mem_free += task.estimated_need_cpu_mem
            task_gpu_id:int = task.gpu_id
            if task_gpu_id != -1:
                self._sys_info.gpu_infos[task_gpu_id].n_cores_free += task.need_gpu_cores
                self._sys_info.gpu_infos[task_gpu_id].mem_free += task.need_gpu_mem

    def _estimate_need_cpu_cores(self, task):
        return task.need_cpu_cores
    
    def _estimate_need_gpu_cores(self, task, gpu_id):
        return task.need_gpu_cores
    
    def _estimate_need_cpu_mem(self, task):
        return (1 - task.modules_overlap_ratio) * task.need_cpu_mem

    def _put_task(self, task:Task)->None:
        self._take_resource(task)
        worker:InterpreterWorker = task.worker
        worker.is_working = True
        worker.imported_modules.update(task.module_deps)
        task.future.set_running_or_notify_cancel()
        worker.task_queue.put(task.info())

    def _add_worker(self)->InterpreterWorker:
        from .interpreterworker import InterpreterWorker

        worker = InterpreterWorker(
            len(self._workers), self._result_queue,
            initializer=self._initializer,
            initargs=self._initargs,
            initkwargs=self._initkwargs,
            torch_cuda_available=self._torch_cuda_available
        )
        self._workers.append(worker)
        return worker



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Callable, Any, Dict, Tuple, Set

from ..worker import Worker

if TYPE_CHECKING:
    from concurrent.interpreters import Queue


class InterpreterWorker(Worker):

    def __init__(
        self, index:int,
        result_queue:Queue[Optional[Tuple[str, bool, Any, int]]],
        initializer:Optional[Callable[..., Any]],
        initargs:Tuple[Any, ...],
        initkwargs:Optional[Dict[str, Any]],
        torch_cuda_available:bool
    ):
        Worker.__init__(self, index, initializer, initargs, initkwargs)

        import concurrent.interpreters as interpreters

        if torch_cuda_available:
            self.change_device_cmd_queue:Optional[Queue[Optional[str]]] = interpreters.create_queue()
        else:
            self.change_device_cmd_queue:Optional[Queue[Optional[str]]] = None

        self.result_queue:Queue[Optional[Tuple[str, bool, Any, int]]] = result_queue
        self.task_queue:Queue[Optional[Tuple[str, Callable[..., Any], Tuple[Any, ...], Dict[str, Any]]]] = interpreters.create_queue()

        self.start()

    def change_device(self, device:str)->None:
        if self.change_device_cmd_queue is not None:
            self.change_device_cmd_queue.put(device)

    def stop(self)->None:
        self.task_queue.put(None)

    def start(self):
        import concurrent.interpreters as interpreters
        from concurrent.interpreters import Interpreter

        self.interp:Interpreter = interpreters.create()
        self.thread = self.interp.call_in_thread(
            InterpreterWorker.run,
            self.task_queue, self.result_queue, self.change_device_cmd_queue,
            initializer=self.initializer,
            initargs=self.initargs,
            initkwargs=self.initkwargs
        )

    def restart(self)->None:
        self.task_queue.put(None)
        self.thread.join()
        self.interp.close()
        self.n_finished_tasks:int = 0
        self.imported_modules.clear()
        self.start()

    def join(self)->None:
        self.thread.join()
        self.interp.close()



from __future__ import annotations
import threading
from queue import SimpleQueue
from typing import TYPE_CHECKING, Optional, Callable, Any, Dict, Tuple

from ..worker import Worker
from ..utils import _set_best_device


if TYPE_CHECKING:
    from ..task import Task


class ThreadWorker(Worker):

    def __init__(
        self, index:int, name_prefix:str,
        result_queue:SimpleQueue[Tuple[str, bool, Any]],
        initializer:Optional[Callable[..., Any]],
        initargs:Tuple[Any, ...],
        initkwargs:Optional[Dict[str, Any]]
    ):
        Worker.__init__(self, index, initializer=initializer, initargs=initargs, initkwargs=initkwargs)
        self.name_prefix:str = name_prefix
        self.task_queue:SimpleQueue[Task] = SimpleQueue()
        self.result_queue:SimpleQueue[Tuple[str, bool, Any]] = result_queue

        self.start()

    def change_device(self, device:str)->None:
        _set_best_device(device, self.thread.ident)

    def start(self)->None:
        self.thread = threading.Thread(
            target=ThreadWorker.run,
            name=f"{self.name_prefix}{self.index}",
            args=(self.task_queue, self.result_queue, self.initializer, self.initargs, self.initkwargs),
            daemon=True
        )
        self.thread.start()

    def stop(self)->None:
        self.task_queue.put(None)

    def join(self)->None:
        self.thread.join()

    def restart(self)->None:
        self.task_queue.put(None)
        self.thread.join()
        self.n_finished_tasks:int = 0
        self.imported_modules.clear()
        self.start()

    @staticmethod
    def run(
        task_queue:SimpleQueue[Task],
        result_queue:SimpleQueue[Tuple[str, bool, Any]],
        initializer:Optional[Callable[..., Any]],
        initargs:Tuple[Any, ...],
        initkwargs:Optional[Dict[str, Any]]
    ):
        if initializer is not None:
            initializer(*initargs, **initkwargs)

        while True:
            task = task_queue.get()
            if task is None:
                break

            _set_best_device(task.device)
            result_queue.put(task.exec())


from __future__ import annotations
import uuid
import sys
from concurrent.futures import Future
from typing import Tuple, Any, Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .worker import Worker


class Task:

    def __init__(self, func, args, kwargs, need_cpu_cores, need_cpu_mem, need_gpu_cores, need_gpu_mem, calculate_module_deps:bool):
        self.id:str = str(uuid.uuid4())
        self.func = func
        self.args:Tuple[Any] = args
        self.kwargs:Dict[str, Any] = kwargs
        self.need_cpu_cores:int = need_cpu_cores
        self.need_cpu_mem:int = need_cpu_mem
        self.need_gpu_cores:int = need_gpu_cores
        self.need_gpu_mem:int = need_gpu_mem
        self.estimated_need_cpu_mem:float = 0.0
        self.modules_overlap_ratio:float = 0.0
        self.module_deps:Dict[str, int] = {}
        
        if calculate_module_deps:
            from .module_deps import module_deps
            self.module_deps:Dict[str, int] = module_deps(sys.modules[func.__module__])
        
        self.device:Optional[str] = None
        self.worker:Worker = None
        self.mem_before_enter:int = 0
        self.future = Future()

    def info(self):
        return self.id, self.device, self.func, self.args, self.kwargs

    @property
    def gpu_id(self)->int:
        if isinstance(self.device, str) and self.device.startswith("cuda:"):
            return int(self.device[len("cuda:"):])
        
        return -1

    def exec(self)->Tuple[str, bool, Any]:
        try:
            result = self.func(*self.args, **self.kwargs)
            success = True
        except BaseException as e:
            result = e
            success = False

        return self.id, success, result


import psutil


class SysInfo:

    def __init__(self):
        self._cpu_cores_total = psutil.cpu_count()
        mem_info = psutil.virtual_memory()
        self._cpu_mem_total = mem_info.total

        self._cpu_cores_used = None
        self._cpu_mem_used = None
        self._gpu_infos = None

    @property
    def cpu_mem_free(self):
        return self._cpu_mem_total - self.cpu_mem_used
    
    @cpu_mem_free.setter
    def cpu_mem_free(self, cpu_mem_free):
        self._cpu_mem_used = self._cpu_mem_total - cpu_mem_free

    @property
    def cpu_mem_used(self):
        if self._cpu_mem_used is None:
            self._cpu_mem_used = psutil.virtual_memory().used

        return self._cpu_mem_used
        
    @cpu_mem_used.setter
    def cpu_mem_used(self, cpu_mem_used):
        self._cpu_mem_used = cpu_mem_used
        
    @property
    def gpu_infos(self):
        from .gpuinfos import GPUInfos

        if self._gpu_infos is None:
            self._gpu_infos = GPUInfos.snapshot("n_cores", "n_cores_used", "mem_total", "mem_used")

        return self._gpu_infos

    @property
    def cpu_cores_free(self):
        return self._cpu_cores_total - self.cpu_cores_used
    
    @cpu_cores_free.setter
    def cpu_cores_free(self, cpu_cores_free):
        self._cpu_cores_used = self._cpu_cores_total - cpu_cores_free

    @property
    def cpu_cores_used(self):
        if self._cpu_cores_used is None:
            self._cpu_cores_used = psutil.cpu_percent() / 100 * self._cpu_cores_total

        return self._cpu_cores_used
        
    @cpu_cores_used.setter
    def cpu_cores_used(self, cpu_cores_used):
        self._cpu_cores_used = cpu_cores_used

    @property
    def cpu_cores_total(self):
        return self._cpu_cores_total
        
    @property
    def cpu_mem_total(self):
        return self._cpu_mem_total

    def update(self):
        self._cpu_core_used = None
        self._cpu_mem_used = None
        self._gpu_infos = None


class DataSize:
    B = 1
    KB = 1024 * B
    MB = 1024 * KB
    GB = 1024 * MB
    TB = 1024 * GB
    PB = 1024 * TB


_has_gil = None

def has_gil()->bool:
    global _has_gil
    if _has_gil is None:
        import sys
        try:
            _has_gil = sys._is_gil_enabled()
        except:
            _has_gil = True

    return _has_gil

def limit_num_single_thread():
    import os
    os.environ["OMP_NUM_THREADS"] = "1"
    os.environ["OPENBLAS_NUM_THREADS"] = "1"
    os.environ["MKL_NUM_THREADS"] = "1"
    os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
    os.environ["NUMEXPR_NUM_THREADS"] = "1"


def move_state_dict_to(obj, device):
    import torch
    import numpy as np

    base_types = (int, float, complex, bool, str, bytes, type(None), np.ndarray, np.bool)

    if isinstance(obj, torch.Tensor):
        return obj.to(device, non_blocking=True)
    elif isinstance(obj, base_types):
        return obj
    elif isinstance(obj, list):
        for i in range(len(obj)):
            sub_obj = obj[i]
            new_sub_obj = move_state_dict_to(sub_obj, device)
            if sub_obj is not new_sub_obj:
                obj[i] = new_sub_obj

        return obj
    elif isinstance(obj, tuple):
        if not all(isinstance(x, base_types) for x in obj):
            return tuple(move_state_dict_to(list(obj), device))
        else:
            return obj
    elif isinstance(obj, dict):
        for k, v in obj.items():
            new_v = move_state_dict_to(v, device)
            if new_v is not v:
                obj[k] = new_v

        return obj
    else:
        raise TypeError(f"{type(obj)} is not supported")


def move_optimizer_to(optimizer, device):
    for state in optimizer.state.values():
        move_state_dict_to(state, device)

    return optimizer


def batched(iterable, chunksize:int):
    import itertools
    iterator = iter(iterable)
    while True:
        batch = list(itertools.islice(iterator, chunksize))
        if not batch:
            break
        yield batch


_best_device_lock = None
_best_device = {}

def _set_best_device(device:str, tid=None)->None:
    import threading

    global _best_device_lock
    if _best_device_lock is None:
        _best_device_lock = threading.Lock()

    if tid is None:
        tid = threading.get_ident()
        
    with _best_device_lock:
        _best_device[tid] = device

def best_device()->str:
    import threading

    global _best_device_lock
    if _best_device_lock is None:
        _best_device_lock = threading.Lock()

    tid = threading.get_ident()
    with _best_device_lock:
        return _best_device[tid]


from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Callable, Any, Dict, Tuple, Set

from ..worker import Worker

if TYPE_CHECKING:
    from multiprocessing.queues import SimpleQueue


class ProcessWorker(Worker):

    def __init__(
        self, index:int, name_prefix:str,
        result_queue:SimpleQueue[Optional[Tuple[str, bool, Any]]], ctx,
        initializer:Optional[Callable[..., Any]],
        initargs:Tuple[Any, ...],
        initkwargs:Optional[Dict[str, Any]],
        use_torch:bool, torch_cuda_available:bool
    ):
        Worker.__init__(
            self, index,
            initializer,
            initargs,
            initkwargs
        )
        if use_torch:
            from torch.multiprocessing.queue import SimpleQueue
        else:
            from multiprocessing.queues import SimpleQueue

        if torch_cuda_available:
            self.change_device_cmd_queue:Optional[SimpleQueue[Optional[str]]] = SimpleQueue(ctx=ctx)
        else:
            self.change_device_cmd_queue:Optional[SimpleQueue[Optional[str]]] = None

        self.ctx = ctx
        self.name_prefix:str = name_prefix
        self.result_queue:SimpleQueue[Optional[Tuple[str, bool, Any]]] = result_queue
        self.task_queue:SimpleQueue[Optional[Tuple[str, Callable[..., Any], Tuple[Any, ...], Dict[str, Any]]]] = SimpleQueue(ctx=ctx)
        self._is_rss_dirty:bool = True
        self._cached_rss:int = 0

        self.start()

    @property
    def cached_rss(self)->int:
        if self._is_working:
            return self.rss

        if self._is_rss_dirty:
            self._cached_rss = self.rss
            self._is_rss_dirty = False

        return self._cached_rss

    @property
    def rss(self)->int:
        try:
            return self.process_info.memory_info().rss
        except:
            return 0

    @property
    def is_working(self)->bool:
        return self._is_working

    @is_working.setter
    def is_working(self, is_working:bool)->None:
        if self._is_working == is_working:
            return
        
        self._is_working = is_working
        self._is_rss_dirty = True

        if is_working:
            Worker._total_working_count += 1
        else:
            Worker._total_working_count -= 1

    def change_device(self, device:str)->None:
        if self.change_device_cmd_queue is not None:
            self.change_device_cmd_queue.put(device)

    def stop(self)->None:
        self.task_queue.put(None)
        self.task_queue.close()

    def start(self):
        import multiprocessing as mp
        import psutil

        self.process:mp.Process = self.ctx.Process(
            target=ProcessWorker.run,
            args=(self.task_queue, self.result_queue, self.change_device_cmd_queue),
            kwargs={"initializer": self.initializer, "initargs": self.initargs, "initkwargs": self.initkwargs},
            name=f"{self.name_prefix}{self.index}",
            daemon=True
        )
        self.process.start()
        self.process_info = psutil.Process(self.process.pid)

    def restart(self)->None:
        self.task_queue.put(None)
        self.process.join()
        self.n_finished_tasks:int = 0
        self.imported_modules.clear()
        self.start()
        
    def terminate(self)->None:
        self.process.terminate()

    def join(self)->None:
        self.process.join()



from __future__ import annotations
from ..pool import Pool
from typing import TYPE_CHECKING, Dict, Tuple, Any, Optional, Callable

if TYPE_CHECKING:
    from ..task import Task
    from .threadworker import ThreadWorker


class ThreadPool(Pool):

    def __init__(
        self, max_workers:int=0, thread_name_prefix:str="ThreadPool.worker:",
        initializer:Optional[Callable[..., Any]]=None,
        initargs:Tuple[Any, ...]=(),
        initkwargs:Optional[Dict[str, Any]]=None,
        *,
        max_tasks_per_child:Optional[int]=None,
        use_torch:bool=False
    ):
        from queue import SimpleQueue

        Pool.__init__(
            self, max_workers=max_workers,

            initializer=initializer,
            initargs=initargs,
            initkwargs=initkwargs,

            result_queue_cls=SimpleQueue,

            max_tasks_per_child=max_tasks_per_child,
            use_torch=use_torch,
            need_module_deps=False
        )

        self._thread_name_prefix:str = thread_name_prefix
        self.__max_used_cpu_cores = None
        self.__max_used_gpu_cores = {}

    def _max_used_cpu_cores(self)->int:
        if self.__max_used_cpu_cores is not None:
            return self.__max_used_cpu_cores

        max_used_cores = 0
        for task in self._tasks.values():
            if task.worker is None or not task.worker.is_working:
                continue

            if task.need_cpu_cores > max_used_cores:
                max_used_cores = task.need_cpu_cores

        self.__max_used_cpu_cores = max_used_cores
        return max_used_cores
    
    def _max_used_gpu_cores(self, gpu_id:int)->int:
        if gpu_id in self.__max_used_gpu_cores:
            return self.__max_used_gpu_cores[gpu_id]

        max_used_cores = 0
        for task in self._tasks.values():
            if task.worker is None or not task.worker.is_working or task.gpu_id != gpu_id:
                continue

            if task.need_gpu_cores > max_used_cores:
                max_used_cores = task.need_gpu_cores

        self.__max_used_gpu_cores[gpu_id] = max_used_cores
        return max_used_cores

    def _has_gil(self)->bool:
        from ..utils import has_gil
        return has_gil()

    def _take_resource(self, task:Task)->None:
        with self._sys_info_lock:
            if not self._has_gil():
                self._sys_info.cpu_cores_free -= task.need_cpu_cores
            else:
                max_used_cpu_cores = self._max_used_cpu_cores()
                if task.need_cpu_cores > max_used_cpu_cores:
                    self.__max_used_cpu_cores = task.need_cpu_cores
                    self._sys_info.cpu_cores_free -= (task.need_cpu_cores - max_used_cpu_cores)

            self._sys_info.cpu_mem_free -= task.need_cpu_mem
            task_gpu_id:int = task.gpu_id
            if task_gpu_id != -1:
                if not self._has_gil():
                    self._sys_info.gpu_infos[task_gpu_id].n_cores_free -= task.need_gpu_cores
                else:
                    max_used_gpu_cores = self._max_used_gpu_cores(task_gpu_id)
                    if task.need_gpu_cores > max_used_gpu_cores:
                        self.__max_used_gpu_cores[task_gpu_id] = task.need_gpu_cores
                        self._sys_info.gpu_infos[task_gpu_id].n_cores_free -= (task.need_gpu_cores - max_used_gpu_cores)

                self._sys_info.gpu_infos[task_gpu_id].mem_free -= task.need_gpu_mem

    def _release_resource(self, task:Task)->None:
        with self._sys_info_lock:
            if not self._has_gil():
                self._sys_info.cpu_cores_free += task.need_cpu_cores
            else:
                max_used_cpu_cores = self._max_used_cpu_cores()
                if task.need_cpu_cores >= max_used_cpu_cores:
                    self.__max_used_cpu_cores = None
                    max_used_cpu_cores = self._max_used_cpu_cores()
                    self._sys_info.cpu_cores_free += (task.need_cpu_cores - max_used_cpu_cores)
                
            self._sys_info.cpu_mem_free += task.need_cpu_mem
            task_gpu_id:int = task.gpu_id
            if task_gpu_id != -1:
                if not self._has_gil():
                    self._sys_info.gpu_infos[task_gpu_id].n_cores_free += task.need_gpu_cores
                else:
                    max_used_gpu_cores = self._max_used_gpu_cores(task_gpu_id)
                    if task.need_gpu_cores >= max_used_gpu_cores:
                        self.__max_used_gpu_cores[task_gpu_id] = None
                        max_used_gpu_cores = self._max_used_gpu_cores(task_gpu_id)
                        self._sys_info.gpu_infos[task_gpu_id].n_cores_free += (task.need_gpu_cores - max_used_gpu_cores)

                self._sys_info.gpu_infos[task_gpu_id].mem_free += task.need_gpu_mem

    def _estimate_need_gpu_cores(self, task:Task, gpu_id:int)->int:
        if self._has_gil():
            return max(0, task.need_gpu_cores - self._max_used_gpu_cores(gpu_id))
        else:
            return task.need_gpu_cores

    def _estimate_need_cpu_cores(self, task:Task)->int:
        if self._has_gil():
            return max(0, task.need_cpu_cores - self._max_used_cpu_cores())
        else:
            return task.need_cpu_cores
        
    def _estimate_need_cpu_mem(self, task:Task)->int:
        return task.need_cpu_mem

    def _put_task(self, task:Task)->None:
        self._take_resource(task)
        worker:ThreadWorker = task.worker
        worker.is_working = True
        task.future.set_running_or_notify_cancel()
        worker.task_queue.put(task)

    def _add_worker(self)->ThreadWorker:
        from .threadworker import ThreadWorker

        worker = ThreadWorker(
            len(self._workers), self._thread_name_prefix,
            self._result_queue,
            initializer=self._initializer,
            initargs=self._initargs,
            initkwargs=self._initkwargs
        )
        self._workers.append(worker)
        return worker



from __future__ import annotations
import pynvml
import weakref
import uuid
import threading
from typing import List, Optional, Iterator


class GPUInfoSnapshot:

    def __init__(self):
        self.id:Optional[int] = 0
        self.name:Optional[str] = None
        self.uuid:Optional[uuid.UUID] = None
        self.serial:Optional[str] = None
        self.driver_version:Optional[str] = None
        self.mem_total:Optional[int] = None
        self.mem_used:Optional[int] = None
        self.load:Optional[float] = None
        self.temperature:Optional[int] = None
        self.display_active:Optional[bool] = None
        self.display_mode:Optional[bool] = None
        self.n_cores:Optional[int] = None
        self.n_cores_used:Optional[int] = None

    @property
    def device(self)->str:
        return f"cuda:{self.id}"
    
    @property
    def mem_free(self)->int:
        return self.mem_total - self.mem_used
    
    @mem_free.setter
    def mem_free(self, mem_free:int)->None:
        self.mem_used = self.mem_total - mem_free
    
    @property
    def n_cores_free(self)->int:
        return self.n_cores - self.n_cores_used
    
    @n_cores_free.setter
    def n_cores_free(self, n_cores_free:int)->None:
        self.n_cores_used = self.n_cores - n_cores_free


class GPUInfo:

    _lock = threading.Lock()

    def __init__(self, index:int) -> None:
        self._id:int = index
        with self._lock:
            self._handle = pynvml.nvmlDeviceGetHandleByIndex(index)

        self._name:Optional[str] = None
        self._uuid:Optional[uuid.UUID] = None
        self._serial:Optional[str] = None
        self._driver_version:Optional[str] = None
        self._mem_info = None
        self._mem_total:Optional[int] = None
        self._mem_used:Optional[int] = None
        self._load:Optional[float] = None
        self._temperature:Optional[int] = None
        self._display_active:Optional[bool] = None
        self._display_mode:Optional[bool] = None
        self._n_cores:Optional[int] = None
        self._n_cores_used:Optional[int] = None
        
    def update(self):
        self._mem_info = None
        self._mem_used:Optional[int] = None
        self._load:Optional[float] = None
        self._temperature:Optional[int] = None
        self._display_active:Optional[bool] = None
        self._display_mode:Optional[bool] = None
        self._n_cores_used:Optional[int] = None

    @property
    def id(self)->int:
        return self._id

    @property
    def device(self)->str:
        return f"cuda:{self.id}"

    @property
    def name(self)->str:
        if self._name is None:
            with self._lock:
                name_bytes = pynvml.nvmlDeviceGetName(self._handle)

            self._name = name_bytes.decode('utf-8') if isinstance(name_bytes, bytes) else name_bytes

        return self._name
    
    @property
    def uuid(self)->uuid.UUID:
        if self._uuid is None:
            with self._lock:
                uuid_bytes = pynvml.nvmlDeviceGetUUID(self._handle)

            self._uuid = uuid.UUID(bytes=uuid_bytes)

        return self._uuid
    
    @property
    def serial(self)->str:
        if self._serial is None:
            with self._lock:
                serial_bytes = pynvml.nvmlDeviceGetSerial(self._handle)

            self._serial = serial_bytes.decode('utf-8') if isinstance(serial_bytes, bytes) else serial_bytes

        return self._serial
    
    @property
    def driver_version(self)->str:
        if self._driver_version is None:
            with self._lock:
                driver_bytes = pynvml.nvmlSystemGetDriverVersion()

            self._driver_version = driver_bytes.decode('utf-8') if isinstance(driver_bytes, bytes) else driver_bytes

        return self._driver_version
    
    @property
    def n_cores(self)->int:
        if self._n_cores is None:
            with self._lock:
                self._n_cores = pynvml.nvmlDeviceGetNumGpuCores(self._handle)

        return self._n_cores
    
    @property
    def n_cores_free(self)->int:
        self._update_load()
        return self._n_cores - self._n_cores_used
    
    @property
    def n_cores_used(self)->int:
        self._update_load()
        return self._n_cores_used
    
    @property
    def mem_total(self)->int:
        if self._mem_total is None:
            if self._mem_info is None:
                with self._lock:
                    self._mem_info = pynvml.nvmlDeviceGetMemoryInfo(self._handle)

            self._mem_total = self._mem_info.total

        return self._mem_total
    
    @property
    def mem_used(self)->int:
        if self._mem_used is None:
            if self._mem_info is None:
                with self._lock:
                    self._mem_info = pynvml.nvmlDeviceGetMemoryInfo(self._handle)

            self._mem_used = self._mem_info.used

        return self._mem_used
    
    @property
    def mem_free(self)->int:
        return self.mem_total - self.mem_used
    
    @property
    def mem_util(self)->float:
        return self.mem_used / self.mem_total
    
    def _update_load(self)->None:
        if self._load is None:
            with self._lock:
                try:
                    util = pynvml.nvmlDeviceGetUtilizationRates(self._handle)
                except:
                    pynvml.nvmlShutdown()
                    pynvml.nvmlInit()
                    self._handle = pynvml.nvmlDeviceGetHandleByIndex(self._id)
                    util = pynvml.nvmlDeviceGetUtilizationRates(self._handle)

            self._load = util.gpu / 100.0
            self._n_cores_used = int(self._load * self.n_cores)

    @property
    def load(self)->float:
        self._update_load()
        return self._load
    
    @property
    def temperature(self)->int:
        if self._temperature is None:
            with self._lock:
                self._temperature = pynvml.nvmlDeviceGetTemperature(self._handle, pynvml.NVML_TEMPERATURE_GPU)

        return self._temperature

    @property
    def display_mode(self)->bool:
        if self._display_mode is None:
            with self._lock:
                self._display_mode = bool(pynvml.nvmlDeviceGetDisplayMode(self._handle))

        return self._display_mode
    
    @property
    def display_active(self)->bool:
        if self._display_active is None:
            with self._lock:
                self._display_active = bool(pynvml.nvmlDeviceGetDisplayActive(self._handle))

        return self._display_active
    
    def __repr__(self)->str:
        return f"GPUInfo(id={self.id}, name='{self.name}')"
    
    def snapshot(self, *args)->GPUInfoSnapshot:
        snapshot = GPUInfoSnapshot()
        snapshot.id = self.id

        if not args or "name" in args:
            snapshot.name = self.name

        if not args or "uuid" in args:
            snapshot.uuid = self.uuid

        if not args or "serial" in args:
            snapshot.serial = self.serial

        if not args or "driver_version" in args:
            snapshot.driver_version = self.driver_version

        if not args or "mem_total" in args or "mem_free" in args or "mem_used" in args:
            snapshot.mem_total = self.mem_total

        if not args or "mem_used" in args or "mem_free" in args:
            snapshot.mem_used = self.mem_used

        if not args or "load" in args:
            snapshot.load = self.load

        if not args or "temperature" in args:
            snapshot.temperature = self.temperature

        if not args or "display_active" in args:
            snapshot.display_active = self.display_active

        if not args or "display_mode" in args:
            snapshot.display_mode = self.display_mode

        if not args or "n_cores" in args or "n_cores_free" in args or "n_cores_used" in args:
            snapshot.n_cores = self.n_cores

        if not args or "n_cores_used" in args or "n_cores_free" in args:
            snapshot.n_cores_used = self.n_cores_used

        return snapshot


class NVMLConstructor:

    def __init__(self):
        pynvml.nvmlInit()
        self.finalizer = weakref.finalize(self, pynvml.nvmlShutdown)


class MetaGPUInfos(type):

    __has_gpu:bool = True
    __nvml_constructor:Optional[NVMLConstructor] = None
    __n_devices:Optional[int] = None
    __gpu_infos:Optional[List[GPUInfo]] = None

    @staticmethod
    def __init()->None:
        if not MetaGPUInfos.__has_gpu:
            return
        
        if MetaGPUInfos.__nvml_constructor is None:
            try:
                MetaGPUInfos.__nvml_constructor = NVMLConstructor()
            except:
                MetaGPUInfos.__has_gpu = False

    def __len__(self)->int:
        if not MetaGPUInfos.__has_gpu:
            return 0

        if MetaGPUInfos.__n_devices is None:
            MetaGPUInfos.__init()
            MetaGPUInfos.__n_devices = pynvml.nvmlDeviceGetCount()

        return MetaGPUInfos.__n_devices

    def __getitem__(self, index:int)->GPUInfo:
        if index >= len(self):
            raise IndexError(f"Index {index} out of range")

        if MetaGPUInfos.__gpu_infos is None:
            MetaGPUInfos.__gpu_infos = []
            for i in range(len(self)):
                MetaGPUInfos.__gpu_infos.append(GPUInfo(i))

        return MetaGPUInfos.__gpu_infos[index]
    
    def __iter__(self)->Iterator[GPUInfo]:
        if MetaGPUInfos.__gpu_infos is None:
            MetaGPUInfos.__gpu_infos = []
            for i in range(len(self)):
                MetaGPUInfos.__gpu_infos.append(GPUInfo(i))

        return iter(MetaGPUInfos.__gpu_infos)


class GPUInfos(metaclass=MetaGPUInfos):
    
    @staticmethod
    def update()->None:
        for gpu in GPUInfos:
            gpu.update()

    @staticmethod
    def snapshot(*args)->List[GPUInfoSnapshot]:
        result = []
        for gpu in GPUInfos:
            gpu.update()
            result.append(gpu.snapshot(*args))

        return result



import sys
import warnings
import functools
import itertools
from types import ModuleType
from typing import Iterable, Set


def _good_module_name(module_name:str, module_names:Set[str])->bool:
    return (
        module_name and 
        module_name not in module_names and
        not module_name.startswith('builtins') and
        module_name in sys.modules
    )

def _module_deps(module:ModuleType)->Set[str]:
    module_names = {module.__name__}
    stack = [module]

    while stack:
        current_module = stack.pop()
        
        for obj in current_module.__dict__.values():
            candidate_name = ""
            if isinstance(obj, ModuleType) and hasattr(obj, '__name__'):
                candidate_name = obj.__name__
            
            elif hasattr(obj, '__module__'):
                candidate_name = obj.__module__
            
            elif hasattr(obj, '__class__') and hasattr(obj.__class__, '__module__'):
                candidate_name = obj.__class__.__module__

            if _good_module_name(candidate_name, module_names):
                sub_module = sys.modules[candidate_name]
                module_names.add(candidate_name)
                stack.append(sub_module)
    
    return module_names


@functools.lru_cache(maxsize=None)
def module_deps(module:ModuleType)->Set[str]:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return _module_deps(module)
